/*
 * lc89075wa_dir_core.c - DIR chip driver over SPI port
 *
 * Copyright 2017, 2018 Sony Corporation

 *  This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/uaccess.h>


#include <linux/gpio.h>
#include <linux/interrupt.h>
#include <linux/spi/spi.h>
#include <mach/mt_spi.h>
#include <mach/mt_gpio.h>
#include <mach/eint.h>
#include <cust_eint.h>
#include <cust_gpio_usage.h>
#include <linux/time.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/wait.h>
#include <linux/sched.h>

#include <linux/spi/lc89075wa_dir.h>
#include "lc89075wa_dir_core.h"
#include "mt8590-afe.h"

#ifdef CONFIG_HAS_EARLYSUSPEND
#include <linux/earlysuspend.h>
#endif

#define LC89075WA_DIR_MODE_READ     0
#define LC89075WA_DIR_MODE_WRITE    1
#define LC89075WA_DIR_BIT_READ      0x80
#define LC89075WA_DIR_BIT_WRITE     0x00
#define DIR_ADDR_REG_INT_STATUS     0xF
#define DIR_SIZE_SETUP_REGS         0xF
#define LC89075WA_DIR_XMODE_STABLE_TIME   200
#define DIR_ERRF_DETECT             0x1

/* decode DIR register code to sample rate in integer [Hz]
 * see DIR register 10h description on LC89075WA chip data sheet
 * read the value of this array at offset RXFSC[3:0] you will get the
   samplerate number
*/
static uint32_t fs_code_2_value[16] = {
	 44100,      0,  48000, 32000,
	     0,      0,      0,     0,
	 88200,      0,  96000, 64000,
	176400, 128000, 192000,     0
};

/* decode bit width code in S/PDIF channel status to bit width in [bit]
 * see S/PDIF specification(IEC 6958-3) or channel status infromation
 * read the value of this array at offset RXCS[35:32] (DIR register 15h)
 * you will get the bit width number
 */
static uint32_t bit_cs_2_width[16] = {
	 0,  0,
	16, 20,
	18, 22,
	 0,  0,
	19, 23,
	20, 24,
	17, 21,
	 0,  0
};

struct lc89075wa_dir_chr_dev_t {
	dev_t id;
	struct cdev cdev;
	struct class *class;
	struct device *dev;
};

struct lc89075wa_driver_data {
	dev_t                  devt;
	spinlock_t             spi_lock;
	struct spi_device      *spi;
	struct mutex           buf_lock;
	unsigned               users;

	struct completion      comp_intr_occur;
	struct task_struct     *kthread_task_read_intr_status;
	wait_queue_head_t      qh_notify_all;
	struct mutex           intr_notify_lock;
	struct mutex           notify_regs_lock;
	int                    retval_last_transfer;

	int                    suspend_flag;

	int                    register_saved;
	uint8_t                register_save_vals[DIR_SIZE_SETUP_REGS];
	struct dir_intr_status last_intr_status;
	int                    chip_has_enabled;
	int                    clk_err_flg;
#ifdef CONFIG_HAS_EARLYSUSPEND
	struct early_suspend   es;
#endif
	bool has_es;
};

/* keep interrupt status registers for notifying to each processes */
struct notify_registers_list {
	struct dir_intr_status intr_status;
	struct list_head       list;
};

/* notify information for each processes */
/* they can be linked with filep's */
struct intr_notify_list {
	struct file                  *filp;
	struct lc89075wa_driver_data *ddata;
	int                          exit_flag;
	int                          notify_list_inited;
	struct notify_registers_list notify_registers_list_head;
	struct list_head             list;
};
static struct intr_notify_list intr_notify_list_head;

static struct lc89075wa_dir_chr_dev_t *lc89075wa_dir_chr_dev;
static struct lc89075wa_driver_data *g_ddata;

static unsigned bufsiz = 4096;
module_param(bufsiz, uint, S_IRUGO);
MODULE_PARM_DESC(bufsiz, "data bytes in biggest supported SPI message");

#define DEBUG_DURATION_ENABLE 0
#if !DEBUG_DURATION_ENABLE	/* release */
static void debug_get_time(int pos) {}
static void debug_show_time(void) {}
#else /* debug: to know how long it takes to handle interrupts */
#define DEBUG_TIME_STR_MAX_BYTES     2000
#define DEBUG_TIME_LOG_MAX_SAMPLES   200
struct debug_time_info {
	struct timeval time;
	int pos;
};
static struct debug_time_info time_info[DEBUG_TIME_LOG_MAX_SAMPLES];
static int timepos;
static void debug_get_time(int pos)
{
	if (timepos < DEBUG_TIME_LOG_MAX_SAMPLES) {
		do_gettimeofday(&time_info[timepos].time);
		time_info[timepos].pos = pos;
		++timepos;
	}
}

static void debug_show_time(void)
{
	int cnt;
	char msg[DEBUG_TIME_STR_MAX_BYTES];
	int duration;

	if (timepos < 2) {
		timepos = 0;
		return;
	}

	msg[0] = 0;
	for (cnt = 0; cnt < (timepos - 1); ++cnt) {
		duration = (time_info[cnt + 1].time.tv_sec
			- time_info[cnt].time.tv_sec) * 1000000
			+ (time_info[cnt + 1].time.tv_usec
			- time_info[cnt].time.tv_usec);
		if (cnt && time_info[cnt].pos == 1) {
			strlcat(msg, "\n", DEBUG_TIME_STR_MAX_BYTES);
		}
		snprintf(msg + strlen(msg),
			DEBUG_TIME_STR_MAX_BYTES - strlen(msg),
			"(%d.%03d)%d", duration / 1000, duration % 1000,
			time_info[cnt + 1].pos);
	}
	dev_info(&g_ddata->spi->dev , "\nintr-du: 1%s[ms]\n", msg);
	timepos = 0;
}
#endif

/* The method of transferring SPI data that using spi_***() APIs
   are based on kernel/drivers/misc/mediatek/spi/mt8590/spi-dev.c
*/
static int lc89075wa_dir_send_command(
	struct spi_device *spi,
	int mode,
	uint8_t addr,
	uint8_t *wr_ptr,
	uint8_t *rd_ptr)
{
	struct spi_transfer xfer;
	uint8_t tx_buffer[2];
	uint8_t rx_buffer[2];
	int ret;
	struct spi_message msg;

	spi_message_init(&msg);


	xfer.tx_buf = tx_buffer;
	xfer.rx_buf = rx_buffer;
	xfer.len = 2;

	if (mode == LC89075WA_DIR_MODE_READ) {
		tx_buffer[0] = LC89075WA_DIR_BIT_READ | addr;
		tx_buffer[1] = 0;
	} else {
		tx_buffer[0] = LC89075WA_DIR_BIT_WRITE | addr;
		if (wr_ptr) {
			tx_buffer[1] = *wr_ptr;
		} else {
			dev_err(&spi->dev,
				"pointer to write data is zero.\n");
			tx_buffer[1] = 0;
		}
	}

	spi_message_add_tail(&xfer, &msg);
	ret = spi_sync(spi, &msg);
	dev_dbg(&spi->dev,
		"spi transfer: ret=%d  tx:0x%02X 0x%02x  rx:0x%02X 0x%02X\n",
		ret, tx_buffer[0], tx_buffer[1], rx_buffer[0], rx_buffer[1]);
	if (ret < 0) {
		dev_warn(&spi->dev, "Message transfer err:%d\n", ret);
		return -1;
	} else {
		if (mode == LC89075WA_DIR_MODE_READ && rd_ptr) {
			*rd_ptr = rx_buffer[1];
		}
		return 0;
	}
}

static int lc89075wa_dir_send_command_multi(
	struct spi_device *spi,
	int mode,
	uint8_t addr,
	uint8_t *wr_ptr,
	uint8_t *rd_ptr,
	int len)
{
	struct spi_transfer xfer;
	uint8_t tx_buffer[LC89075WA_DIR_IOCTL_MAX_BYTES];
	uint8_t rx_buffer[LC89075WA_DIR_IOCTL_MAX_BYTES];
	int ret;
	struct spi_message msg;

	if (len > LC89075WA_DIR_IOCTL_MAX_BYTES) {
		return -2;
	}

	spi_message_init(&msg);


	xfer.tx_buf = tx_buffer;
	xfer.rx_buf = rx_buffer;
	xfer.len = len + 1;

	memset(tx_buffer, 0, sizeof(tx_buffer));
	if (mode == LC89075WA_DIR_MODE_READ) {
		tx_buffer[0] = LC89075WA_DIR_BIT_READ | addr;
	} else {
		tx_buffer[0] = LC89075WA_DIR_BIT_WRITE | addr;
		if (wr_ptr) {
			memcpy(tx_buffer + 1, wr_ptr, len);
		} else {
			dev_err(&spi->dev,
				"pointer to write data is zero.\n");
		}
	}

	spi_message_add_tail(&xfer, &msg);
	ret = spi_sync(spi, &msg);
	dev_dbg(&spi->dev,
		"spi transfer: ret=%d  tx:0x%02X 0x%02x  rx:0x%02X 0x%02X\n",
		ret, tx_buffer[0], tx_buffer[1], rx_buffer[0], rx_buffer[1]);
	if (ret < 0) {
		dev_warn(&spi->dev, "Message transfer err:%d\n", ret);
		return -1;
	} else {
		if (mode == LC89075WA_DIR_MODE_READ && rd_ptr) {
			memcpy(rd_ptr, rx_buffer + 1, len);
		}
		return 0;
	}
}

static void lc89075wa_dir_init_notify_list(
	struct intr_notify_list *notify_list)
{
	struct lc89075wa_driver_data *ddata;

	ddata = notify_list->ddata;
	if (!notify_list->notify_list_inited) {
		INIT_LIST_HEAD(&notify_list->notify_registers_list_head.list);

		mutex_lock(&ddata->intr_notify_lock);
		list_add_tail(&notify_list->list, &intr_notify_list_head.list);
		notify_list->notify_list_inited = 1;
		mutex_unlock(&ddata->intr_notify_lock);
	}
}


/*
 * lc89075wa_dir driver file operations
 */

static int lc89075wa_dir_open(struct inode *inode, struct file *filp)
{
	struct intr_notify_list *notify_list;

	if (!g_ddata) {
		pr_err("lc89075wa_dir_core.ko: open failed - "
			"device driver is not ready\n");
		return -ENODEV;
	}

	/* keep spi driver information in filp */
	notify_list = kzalloc(sizeof(*notify_list), GFP_KERNEL);
	if (!notify_list) {
		return -ENOMEM;
	}

	g_ddata->users++;
	notify_list->ddata = g_ddata;
	notify_list->filp = filp;
	filp->private_data = notify_list;
	nonseekable_open(inode, filp);

	return 0;
}

static int lc89075wa_dir_close(struct inode *inode, struct file *filp)
{
	int status = 0;
	struct intr_notify_list *notify_list;
	struct lc89075wa_driver_data *ddata;

	notify_list = (struct intr_notify_list *)filp->private_data;
	ddata = notify_list->ddata;

	if (notify_list->notify_list_inited) {
		int hit = 0;
		struct intr_notify_list *st_notify;
		struct list_head *list, *prev;
		struct notify_registers_list *st_regs;

		mutex_lock(&ddata->intr_notify_lock);
		list_for_each(list, &intr_notify_list_head.list) {
			st_notify = list_entry(list, struct intr_notify_list,
				list);
			dev_dbg(&ddata->spi->dev,
				"search filp=0x%p   delete 0x%p\n",
				st_notify->filp, filp);
			if (st_notify->filp == filp) {
				hit = 1;
				break;
			}
		}
		if (hit) {
			dev_dbg(&ddata->spi->dev, "removed filp=0x%p\n",
				st_notify->filp);
			list_del(list);
			/* struct intr_norify_list = filp->private_data is
			   freed below */
		} else {
			dev_err(&ddata->spi->dev,
				"could not find notify_list to remove\n");
		}
		mutex_unlock(&ddata->intr_notify_lock);

		/* remove all interrupt status register data,
		   if they are in the list */
		mutex_lock(&ddata->notify_regs_lock);
		list_for_each_prev_safe(list, prev,
			&notify_list->notify_registers_list_head.list) {
			st_regs = list_entry(list,
				struct notify_registers_list, list);
			list_del(list);
			kfree(st_regs);
		}
		mutex_unlock(&ddata->notify_regs_lock);
	}

	if (!filp->private_data) {
		dev_err(&ddata->spi->dev, "spi device was already closed\n");
		return -EBADF;
	}

	kfree(filp->private_data); /* struct intr_notify_list */
	filp->private_data = NULL;
	ddata->users--;
	return status;
}

/* wait INTB assertion and read status register */
static ssize_t lc89075wa_dir_read(
	struct file *filp,
	char __user *buf,
	size_t size,
	loff_t *offset)
{
	struct intr_notify_list *notify_list;
	struct lc89075wa_driver_data  *ddata;
	int ret = 0;

	if (filp->f_flags & O_NONBLOCK) {
		ret = -EAGAIN;
		dev_warn(&g_ddata->spi->dev, "%s : goto fail\n",
			__func__);
		goto fail;
	}

	notify_list = (struct intr_notify_list *)filp->private_data;
	ddata = notify_list->ddata;
	if (!notify_list->notify_list_inited) {
		lc89075wa_dir_init_notify_list(notify_list);
	}

	/* if no interrupt status data have added to the list */
	if (list_empty(&notify_list->notify_registers_list_head.list)) {
		return 0;
	}

	if (ret || notify_list->exit_flag) {
		dev_info(&ddata->spi->dev,
			"woke_up : wait_for_completion_interruptible = %d\n",
			ret);
		if (ret == -ERESTARTSYS) {
			dev_warn(&ddata->spi->dev, "wait_for_completion"
				"(SPI_DIR) stopped with signal ret=%d\n",
				ret);
		} else if (!ret) {
			dev_dbg(&ddata->spi->dev, "wait_for_completion "
				"exited by requested_to_exit\n");
		} else {
			dev_warn(&ddata->spi->dev, "wait_for_completion "
				"exited unexpected condition ret=%d\n", ret);
		}
		mutex_lock(&ddata->intr_notify_lock);
		if (notify_list->exit_flag) {
			notify_list->exit_flag = 0;  /* clear flag */
		}
		mutex_unlock(&ddata->intr_notify_lock);
		goto fail;
	}

	if (ddata->retval_last_transfer) {
		ret = ddata->retval_last_transfer;
		goto fail;
	}
	if (buf) {
		struct notify_registers_list *st_regs;

		mutex_lock(&ddata->notify_regs_lock);
		st_regs = list_entry(
			notify_list->notify_registers_list_head.list.next,
			struct notify_registers_list, list);
		list_del(notify_list->notify_registers_list_head.list.next);
		mutex_unlock(&ddata->notify_regs_lock);

		ret = copy_to_user(buf, &st_regs->intr_status, size);
		kfree(st_regs);
		if (ret) {
			goto fail;
		}
	}
	return size;

fail:
	dev_err(&g_ddata->spi->dev, "%s : return %d\n", __func__, ret);
	return ret;
}

static ssize_t lc89075wa_dir_write(
	struct file *filp,
	const char __user *buf,
	size_t size,
	loff_t *offset)
{
	dev_dbg(&g_ddata->spi->dev, "write\n");
	return 0;
}

static int lc89075wa_dir_set_chip_enable(
	struct lc89075wa_driver_data *ddata,
	struct spi_device *spi,
	struct lc89075wa_dir_ioctl_params *pparam)
{
	int retval = -EINVAL;
	if (pparam->value != LC89075WA_DIR_CHIP_DISABLE &&
	    ddata->register_saved) {
		dev_dbg(&g_ddata->spi->dev, "recover DIR registers\n");
		mt_set_gpio_out(GPIO_DIR_RST_EN_PIN, GPIO_OUT_ONE);
		/* wait until DIR got stable */
		usleep_range(LC89075WA_DIR_XMODE_STABLE_TIME,
			LC89075WA_DIR_XMODE_STABLE_TIME + 10);
		retval = lc89075wa_dir_send_command_multi(spi,
			LC89075WA_DIR_MODE_WRITE, 0,
			ddata->register_save_vals, NULL,
			sizeof(ddata->register_save_vals));

		if (!retval) {
			ddata->chip_has_enabled = 1;
			ddata->register_saved = 0;
			complete(&ddata->comp_intr_occur);
		}
		else
			mt_set_gpio_out(GPIO_DIR_RST_EN_PIN, GPIO_OUT_ZERO);
	} else if (pparam->value == LC89075WA_DIR_CHIP_DISABLE &&
		   !ddata->register_saved) {
		dev_dbg(&g_ddata->spi->dev, "save DIR registers\n");
		retval = lc89075wa_dir_send_command_multi(spi,
			LC89075WA_DIR_MODE_READ, 0,
			NULL, ddata->register_save_vals,
			sizeof(ddata->register_save_vals));
		if (!retval) {
			ddata->register_saved = 1;
			mt_set_gpio_out(GPIO_DIR_RST_EN_PIN, GPIO_OUT_ZERO);
		}
	}
	return retval;
}

static long lc89075wa_dir_ioctl(
	struct file *filp,
	unsigned int cmd,
	unsigned long arg)
{
	int err = 0;
	int retval = 0;
	struct intr_notify_list *notify_list;
	struct lc89075wa_driver_data  *ddata;
	struct spi_device *spi;
	struct lc89075wa_dir_ioctl_params param;
	uint8_t rd_data, wr_data;

	/* guard against device removal before, or while,
	 * we issue this ioctl.
	 */
	notify_list = (struct intr_notify_list *)filp->private_data;
	ddata = notify_list->ddata;
	spin_lock_irq(&ddata->spi_lock);
	spi = spi_dev_get(ddata->spi);
	spin_unlock_irq(&ddata->spi_lock);

	if (spi == NULL)
		return -ESHUTDOWN;

	memset(&param, 0, sizeof(struct lc89075wa_dir_ioctl_params));
	err = copy_from_user(&param, (const void *)arg,
		LC89075WA_DIR_IOCTL_PARAM_BASIC_SIZE);
	if (err) {
		/* some of the data were not able to copy */
		return -EFAULT;
	}

	/* use the buffer lock here for triple duty:
	 *  - prevent I/O (from us) so calling spi_setup() is safe;
	 *  - prevent concurrent SPI_IOC_WR_* from morphing
	 *    data fields while SPI_IOC_RD_* reads them;
	 *  - SPI_IOC_MESSAGE needs the buffer locked "normally".
	 */
	mutex_lock(&ddata->buf_lock);

	switch (param.cmd) {
	case LC89075WA_DIR_IOCTL_WRITE_REG:
		/* check suspend status before starting SPI transfer */
		if (ddata->suspend_flag) {
			retval = -EBUSY;
			break;
		}
		dev_dbg(&spi->dev,
			"ioctl  write address = 0x%X   value 0x%X\n",
			param.addr, param.value);
		wr_data = param.value;

		if (DIR_SIZE_SETUP_REGS <= param.addr) {
			retval = -EINVAL;
			break;
		}
		retval = lc89075wa_dir_send_command(spi,
			LC89075WA_DIR_MODE_WRITE,
			param.addr, &wr_data, NULL);
		if (!retval) {
			ddata->register_save_vals[param.addr] = wr_data;
		}
		break;
	case LC89075WA_DIR_IOCTL_READ_REG:
		/* check suspend status before starting SPI transfer */
		if (ddata->suspend_flag) {
			retval = -EBUSY;
			break;
		}
		dev_dbg(&spi->dev, "ioctl  read  from address = 0x%X\n",
			param.addr);
		retval = lc89075wa_dir_send_command(spi,
			LC89075WA_DIR_MODE_READ,
			param.addr, NULL, &rd_data);
		if (retval) {
			break;
		}
		param.value = rd_data;
		dev_dbg(&spi->dev, "ioctl  read  result 0x%X\n",
			param.value);
		err = copy_to_user((void *)arg, &param,
			sizeof(struct lc89075wa_dir_ioctl_params));
		if (err) {
			retval = -EFAULT;
			break;
		}
		break;
	case LC89075WA_DIR_IOCTL_EXIT_WAIT:
		dev_dbg(&spi->dev, "ioctl request to exit from intr wait\n");
		if (notify_list->notify_list_inited) {
			mutex_lock(&ddata->intr_notify_lock);
			notify_list->exit_flag = TRUE;
			mutex_unlock(&ddata->intr_notify_lock);
			wake_up_interruptible(&ddata->qh_notify_all);
		}
		break;
	case LC89075WA_DIR_IOCTL_SET_CHIP_ENABLE:
		dev_dbg(&spi->dev, "ioctl chip /reset %d(%s)  saved=%d\n",
			param.value, param.value ? "enabled" : "disabled",
			ddata->register_saved);
		retval = lc89075wa_dir_set_chip_enable(ddata, spi,
			&param);
		break;
	case LC89075WA_DIR_IOCTL_SET_ERRF_ENABLE:
	{
		uint8_t new06;

		if (ddata->suspend_flag) {
			retval = -EBUSY;
			break;
		}

		new06 = (param.value ? LC89075WA_DIR_OUT_ERRF_ERRORFLAG :
			LC89075WA_DIR_OUT_ERRF_MPIN4)
			| (ddata->register_save_vals[6] &
			~LC89075WA_DIR_REG_MASK_FLGERR);
		retval = lc89075wa_dir_send_command(spi,
			LC89075WA_DIR_MODE_WRITE,
			LC89075WA_DIR_ADDR_SELECT_OUTPUT, &new06, NULL);
		if (retval) {
			break;
		}

		ddata->register_save_vals[6] = new06;
		break;
	}
	case LC89075WA_DIR_IOCTL_SELECT_INPUT:
	{
		/* set MCKOUT bits to choose one of SPDIF, ADC, GND */
		/* set FLGERR to be ERRF no error */
		/* set FLGOUT to be MUTE none mute */
		uint8_t new06 = (param.value & LC89075WA_DIR_REG_MASK_MCKOUT)
			| ((param.value & LC89075WA_DIR_REG_MASK_MCKOUT) ?
			  LC89075WA_DIR_REG_MASK_FLGERR:0x00)
			| (((param.value & LC89075WA_DIR_REG_MASK_MCKOUT)
			  > LC89075WA_DIR_MCKOUT_ADC) ? LC89075WA_DIR_REG_MASK_FLGOUT:0x00);

		/* set RXDSEL bits to choose SPDIF sources */
		uint8_t new0B = (param.value & LC89075WA_DIR_REG_MASK_RXDSEL)
			| (ddata->register_save_vals[0xB] &
			~LC89075WA_DIR_REG_MASK_RXDSEL);

		if (ddata->suspend_flag) {
			retval = -EBUSY;
			break;
		}

		dev_dbg(&spi->dev,
			"ioctl select input [06]=0x%X  [0B]=0x%X\n",
			new06, new0B);

		/* select SPDIF first then stop I2S or analog */
		retval = lc89075wa_dir_send_command(spi,
			LC89075WA_DIR_MODE_WRITE,
			LC89075WA_DIR_ADDR_SELECT_DIR_INPUT,
			&new0B, NULL);
		if (retval)
			break;

		retval = lc89075wa_dir_send_command(spi,
			LC89075WA_DIR_MODE_WRITE,
			LC89075WA_DIR_ADDR_SELECT_OUTPUT,
			&new06, NULL);
		if (retval) {
			ddata->register_save_vals[0xB] = new0B;
			break;
		}
		ddata->register_save_vals[6] = new06;
		ddata->register_save_vals[0xB] = new0B;
		break;
	}
	case LC89075WA_DIR_IOCTL_GET_INPUT_NUM:
	{
		uint8_t bits_mckout = ddata->register_save_vals[6]
			& LC89075WA_DIR_REG_MASK_MCKOUT;
		uint8_t bits_rxdsel = ddata->register_save_vals[0xB]
			& LC89075WA_DIR_REG_MASK_RXDSEL;
		param.value = bits_rxdsel | bits_mckout;

		err = copy_to_user((void *)arg, &param,
			sizeof(struct lc89075wa_dir_ioctl_params));
		if (err) {
			retval = -EFAULT;
			break;
		}
		break;
	}
	case LC89075WA_DIR_IOCTL_SET_ADC_VOL:
	{
		uint8_t reg08;

		if (ddata->suspend_flag) {
			retval = -EBUSY;
			break;
		}

		if (param.lvalue <= LC89075WA_DIR_ADCVOL_LEVEL_m4_5dB) {
			reg08 = LC89075WA_DIR_ADCVOL_CODE_m4_5dB;
		} else if (param.lvalue <= LC89075WA_DIR_ADCVOL_LEVEL_m3_0dB) {
			reg08 = LC89075WA_DIR_ADCVOL_CODE_m3_0dB;
		} else if (param.lvalue <= LC89075WA_DIR_ADCVOL_LEVEL_m1_5dB) {
			reg08 = LC89075WA_DIR_ADCVOL_CODE_m1_5dB;
		} else if (param.lvalue <= LC89075WA_DIR_ADCVOL_LEVEL_0_0dB) {
			reg08 = LC89075WA_DIR_ADCVOL_CODE_0_0dB;
		} else if (param.lvalue <= LC89075WA_DIR_ADCVOL_LEVEL_1_5dB) {
			reg08 = LC89075WA_DIR_ADCVOL_CODE_1_5dB;
		} else if (param.lvalue <= LC89075WA_DIR_ADCVOL_LEVEL_3_0dB) {
			reg08 = LC89075WA_DIR_ADCVOL_CODE_3_0dB;
		} else if (param.lvalue <= LC89075WA_DIR_ADCVOL_LEVEL_4_5dB) {
			reg08 = LC89075WA_DIR_ADCVOL_CODE_4_5dB;
		} else {
			reg08 = LC89075WA_DIR_ADCVOL_CODE_6_0dB;
		}

		dev_dbg(&spi->dev, "ioctl set ADC volume lvalue=%d  0x%X\n",
			param.lvalue, reg08);
		/* assume: mute settings with 08h register are not changed */
		retval = lc89075wa_dir_send_command(spi,
			LC89075WA_DIR_MODE_WRITE,
			LC89075WA_DIR_ADDR_SET_ADCPGA, &reg08, NULL);
		break;
	}
	case LC89075WA_DIR_IOCTL_GET_PLL_FLAG:
		param.value = (0 != (ddata->last_intr_status.IntrCause &
			LC89075WA_DIR_REG_MASK_OERROR));
		err = copy_to_user((void *)arg, &param,
			sizeof(struct lc89075wa_dir_ioctl_params));
		if (err) {
			retval = -EFAULT;
			break;
		}
		break;
	case LC89075WA_DIR_IOCTL_GET_NONPCM_FLAG:
		param.value = (0 != (ddata->last_intr_status.IntrCause &
			LC89075WA_DIR_REG_MASK_OUNPCM));
		err = copy_to_user((void *)arg, &param,
			sizeof(struct lc89075wa_dir_ioctl_params));
		if (err) {
			retval = -EFAULT;
			break;
		}
		break;
	case LC89075WA_DIR_IOCTL_GET_EMPHASIS_FLAG:
		param.value = (0 != (ddata->last_intr_status.IntrCause
			& LC89075WA_DIR_REG_MASK_OEMPHA));
		err = copy_to_user((void *)arg, &param,
			sizeof(struct lc89075wa_dir_ioctl_params));
		if (err) {
			retval = -EFAULT;
			break;
		}
		break;
	case LC89075WA_DIR_IOCTL_GET_SAMPLE_RATE:
	{
		param.lvalue = fs_code_2_value[
			ddata->last_intr_status.InfoFsType &
			LC89075WA_DIR_REG_MASK_RXFSC];
		err = copy_to_user((void *)arg, &param,
			sizeof(struct lc89075wa_dir_ioctl_params));
		if (err) {
			retval = -EFAULT;
			break;
		}
		break;
	}
	case LC89075WA_DIR_IOCTL_GET_BIT_WIDTH:
		param.value = bit_cs_2_width[
			ddata->last_intr_status.ChStatus[4] &
			LC89075WA_DIR_REG_MASK_BITWIDTH];
		err = copy_to_user((void *)arg, &param,
			sizeof(struct lc89075wa_dir_ioctl_params));
		if (err) {
			retval = -EFAULT;
			break;
		}
		break;
	case LC89075WA_DIR_IOCTL_GET_INTR_STATUS:
		memcpy(&param.intr_status, &ddata->last_intr_status,
			sizeof(param.intr_status));
		dev_dbg(&spi->dev, "ioctl intr_status 0x%02X 0x%02X: "
			"0x%02X 0x%02X 0x%02X 0x%02X 0x%02X\n",
			(int)param.intr_status.IntrCause,
			(int)param.intr_status.InfoFsType,
			(int)param.intr_status.ChStatus[0],
			(int)param.intr_status.ChStatus[1],
			(int)param.intr_status.ChStatus[2],
			(int)param.intr_status.ChStatus[3],
			(int)param.intr_status.ChStatus[4]);
		err = copy_to_user((void *)arg, &param,
			sizeof(struct lc89075wa_dir_ioctl_params));
		if (err) {
			retval = -EFAULT;
			break;
		}
		break;

	/* for test */
	case LC89075WA_DIR_IOCTL_TEST_SUSPEND:
		dev_info(&spi->dev,
			"ioctl to set suspend_flag = %d(set flag only)\n",
			param.value);
		/* check suspend status before starting SPI transfer */
		ddata->suspend_flag = param.value;
		if (!param.value) {
			mt_eint_unmask(CUST_EINT_SPI_DIR_NUM);
		}
		break;

	default:
		retval = -EINVAL;
	}

	mutex_unlock(&ddata->buf_lock);
	spi_dev_put(spi);
	return retval;
}

static void lc89075wa_dir_irq_handler(void)
{
	if (g_ddata) {
		debug_get_time(1);
		complete(&g_ddata->comp_intr_occur);
	}
}

static int lc89075wa_dir_get_push_intr_status(
	struct lc89075wa_driver_data *ddata, struct spi_device *spi)
{
	struct intr_notify_list *st_notify;
	struct list_head *list;
	int ret;

	mutex_lock(&ddata->buf_lock);
	ret = lc89075wa_dir_send_command_multi(spi, LC89075WA_DIR_MODE_READ,
		DIR_ADDR_REG_INT_STATUS, NULL,
		(uint8_t *)&ddata->last_intr_status,
		sizeof(ddata->last_intr_status));
	mutex_unlock(&ddata->buf_lock);

	debug_get_time(3);
	if (ret) {
		return ret;
	}

	if ((ddata->last_intr_status.IntrCause & DIR_ERRF_DETECT) == 0 &&
		ddata->clk_err_flg)
		extdir_clock_recovery();

	ddata->clk_err_flg = (ddata->last_intr_status.IntrCause & DIR_ERRF_DETECT);

	mutex_lock(&ddata->intr_notify_lock);
	list_for_each(list, &intr_notify_list_head.list) {
		struct notify_registers_list *st_regs;
		st_regs = kzalloc(sizeof(struct notify_registers_list),
			GFP_KERNEL);
		if (!st_regs) {
			ret = -ENOMEM;
			break;
		}
		st_notify = list_entry(list, struct intr_notify_list, list);
		memcpy(&st_regs->intr_status, &ddata->last_intr_status,
			sizeof(st_regs->intr_status));
		mutex_lock(&ddata->notify_regs_lock);
		list_add_tail(&st_regs->list,
			&st_notify->notify_registers_list_head.list);
		mutex_unlock(&ddata->notify_regs_lock);
	}
	mutex_unlock(&ddata->intr_notify_lock);
	debug_get_time(4);

	mt_eint_unmask(CUST_EINT_SPI_DIR_NUM);

	wake_up_interruptible(&ddata->qh_notify_all);
	return ret;
}

static int kthread_func_read_intr_status(void *arg)
{
	struct lc89075wa_driver_data *ddata =
		(struct lc89075wa_driver_data *)arg;
	struct spi_device *spi;
	int ret;

	/* guard against device removal before, or while,
	 * we issue this ioctl.
	 */
	spin_lock_irq(&ddata->spi_lock);
	spi = spi_dev_get(ddata->spi);
	spin_unlock_irq(&ddata->spi_lock);

	if (spi == NULL) {
		return -ESHUTDOWN;
	}

	while (!kthread_should_stop()) {
		debug_get_time(6);
		debug_get_time(7);

		dev_dbg(&ddata->spi->dev, "# waiting DIR interrupt\n");
		if (!(ddata->last_intr_status.IntrCause &
			LC89075WA_DIR_REG_MASK_OERROR)) {
			debug_show_time();
		}
		ret = wait_for_completion_interruptible(
			&ddata->comp_intr_occur);
		debug_get_time(2);

		if (ddata->suspend_flag) {
			ret = -EBUSY;
		} else {
			ret = lc89075wa_dir_get_push_intr_status(ddata, spi);
		}
		ddata->retval_last_transfer = ret;
		debug_get_time(5);
	}
	spi_dev_put(spi);

	dev_dbg(&ddata->spi->dev,
		"kthread_func_read_intr_status exited\n");
	return 0;
}

/* block until interrupt occurred and intr status got ready */
static unsigned int lc89075wa_dir_poll(struct file *filp, poll_table *wait)
{
	struct intr_notify_list *notify_list;
	unsigned int mask = 0;

	notify_list = (struct intr_notify_list *)filp->private_data;

	if (!notify_list->notify_list_inited) {
		lc89075wa_dir_init_notify_list(notify_list);
	}

	/* if status register have not read, wait until interrupts are occurred,
	  that would wake this up with qh_norify_all.
	  If interrrupts have already occurred and the status registers have
	  read and kept, go through here */
	if (list_empty(&notify_list->notify_registers_list_head.list)) {
		poll_wait(filp, &notify_list->ddata->qh_notify_all,
			wait);
		debug_get_time(19);
	}

	if (notify_list->exit_flag) {
		mask = POLLERR;
	} else if (list_empty(&notify_list->notify_registers_list_head.list)) {
		mask = 0;
	} else {
		mask = POLLIN | POLLRDNORM;
	}
	return mask;
}

#ifdef CONFIG_HAS_EARLYSUSPEND
static void lc89075wa_dir_spi_early_suspend(struct early_suspend *h)
{
	int retval = -EINVAL;

	pr_debug("spi:lc89075 early_suspend\n");

	mt_eint_registration(CUST_EINT_SPI_DIR_NUM, CUST_EINT_SPI_DIR_TYPE,
		NULL, 0);

	if (g_ddata) {
		g_ddata->suspend_flag = 1;  /* set flag */
		retval = lc89075wa_dir_send_command_multi(g_ddata->spi,
			LC89075WA_DIR_MODE_READ, 0,
			NULL, g_ddata->register_save_vals,
			sizeof(g_ddata->register_save_vals));
		if (!retval)
			g_ddata->register_saved = 1;
	}

	mt_set_gpio_out(GPIO_DIR_RST_EN_PIN, GPIO_OUT_ZERO);
}

static void lc89075wa_dir_spi_late_resume(struct early_suspend *h)
{
	struct lc89075wa_dir_platform_data *pfdata;
	int retval = -EINVAL;

	pr_debug("spi:lc89075 late_resume\n");

	mt_set_gpio_out(GPIO_DIR_RST_EN_PIN, GPIO_OUT_ONE);

	/* wait until DIR got stable */
	usleep_range(LC89075WA_DIR_XMODE_STABLE_TIME,
		LC89075WA_DIR_XMODE_STABLE_TIME + 10);

	if (g_ddata) {
		g_ddata->suspend_flag = 0;  /* clear flag */
		if (!g_ddata->register_saved) {
			/* use default values */
			pfdata = (struct lc89075wa_dir_platform_data *)
				g_ddata->spi->dev.platform_data;
			memcpy(g_ddata->register_save_vals,
				pfdata->reg_init_value,
				sizeof(g_ddata->register_save_vals));
		}
		retval = lc89075wa_dir_send_command_multi(g_ddata->spi,
			LC89075WA_DIR_MODE_WRITE, 0,
			g_ddata->register_save_vals, NULL,
			sizeof(g_ddata->register_save_vals));
		if (!retval) {
			g_ddata->register_saved = 0;
			complete(&g_ddata->comp_intr_occur);
		}
	}

	mt_eint_set_hw_debounce(
		CUST_EINT_SPI_DIR_NUM, CUST_EINT_SPI_DIR_DEBOUNCE_CN);
	mt_eint_registration(CUST_EINT_SPI_DIR_NUM, CUST_EINT_SPI_DIR_TYPE,
		lc89075wa_dir_irq_handler, 0);
}
#endif

/* to link with spi driver */
static int lc89075wa_dir_spi_probe(struct spi_device *spi)
{
	struct lc89075wa_driver_data *ddata;
	int status;
	struct mt_chip_conf *chip_config = NULL;
	struct lc89075wa_dir_platform_data *pfdata;
	struct lc89075wa_dir_ioctl_params param;
	uint8_t *init_regs;

	if (g_ddata) {
		dev_err(&spi->dev, "DIR device has been already probed.\n");
		return -EBUSY;
	}
	/* Allocate driver data */
	ddata = kzalloc(sizeof(*ddata), GFP_KERNEL);
	if (!ddata) {
		return -ENOMEM;
	}

	/* Initialize the driver data */
	ddata->spi = spi;
	spin_lock_init(&ddata->spi_lock);
	mutex_init(&ddata->buf_lock);

	dev_dbg(&spi->dev, "before create_device   lc89075wa_dir%d.%d\n",
		spi->master->bus_num, spi->chip_select);

	chip_config = (struct mt_chip_conf *) spi->controller_data;
	dev_dbg(&spi->dev, "chip_config=0x%X  cpol=%d",
		(unsigned int)chip_config,
		chip_config ? chip_config->cpol : (-1));

	/* minor number=0 */
	ddata->devt = MKDEV(MAJOR(lc89075wa_dir_chr_dev->id), 0);
	spi_set_drvdata(spi, ddata);
	g_ddata = ddata;
	status = 0;

	/* GPIO83(DIR-INTB) interrupt settings */
	init_completion(&ddata->comp_intr_occur);
	mutex_init(&ddata->notify_regs_lock);
	mutex_init(&ddata->intr_notify_lock);

	/* GPIO83 settings to receive INTB from DIR chip */
	mt_set_gpio_mode(GPIO_DIR_INTB_PIN, GPIO_DIR_INTB_PIN_M_EINT);
	mt_set_gpio_dir(GPIO_DIR_INTB_PIN, GPIO_DIR_IN);
	mt_set_gpio_pull_enable(GPIO_DIR_INTB_PIN, GPIO_PULL_ENABLE);
	mt_set_gpio_pull_select(GPIO_DIR_INTB_PIN, GPIO_PULL_DOWN);

	mt_eint_set_hw_debounce(
		CUST_EINT_SPI_DIR_NUM, CUST_EINT_SPI_DIR_DEBOUNCE_CN);
	mt_eint_registration(CUST_EINT_SPI_DIR_NUM, CUST_EINT_SPI_DIR_TYPE,
		lc89075wa_dir_irq_handler, 0);
	ddata->suspend_flag = 0;

	INIT_LIST_HEAD(&intr_notify_list_head.list);
	init_waitqueue_head(&ddata->qh_notify_all);

	ddata->kthread_task_read_intr_status =
		kthread_run(kthread_func_read_intr_status, ddata,
		"lc89075wa_thread_read_status");
	if (IS_ERR(ddata->kthread_task_read_intr_status)) {
		status = (int)ddata->kthread_task_read_intr_status;
		goto fail_thread_run;
	}


	/* set up the pin to work as GPIO that is connected to
	   DIR XMODE(/reset) */
	mt_set_gpio_mode(GPIO_DIR_RST_EN_PIN, GPIO_DIR_RST_EN_PIN_M_GPIO);

	pfdata = (struct lc89075wa_dir_platform_data *)spi->dev.platform_data;
	init_regs = pfdata->reg_init_value;
	dev_dbg(&spi->dev, "platform_data=0x%p []=0x%02X\n",
		init_regs, init_regs ? init_regs[0]:0);
	memcpy(ddata->register_save_vals, init_regs,
		sizeof(ddata->register_save_vals));
	ddata->register_saved = 1;
	ddata->chip_has_enabled = 0;
#ifdef CONFIG_HAS_EARLYSUSPEND
	ddata->es.suspend = lc89075wa_dir_spi_early_suspend;
	ddata->es.resume = lc89075wa_dir_spi_late_resume;
	ddata->es.level = EARLY_SUSPEND_LEVEL_BLANK_SCREEN;
	register_early_suspend(&ddata->es);
	ddata->has_es = TRUE;
#endif

	param.value = TRUE;
	lc89075wa_dir_set_chip_enable(ddata, spi, &param);
	return 0;

fail_thread_run:
	dev_err(&spi->dev, "lc89075wa_dir_spi_probe failed %d\n", status);
	mutex_destroy(&ddata->intr_notify_lock);
	mutex_destroy(&ddata->notify_regs_lock);
	spi_set_drvdata(spi, NULL);
	mutex_destroy(&ddata->buf_lock);
	kfree(ddata);
	g_ddata = NULL;

	return status;
}

static int lc89075wa_dir_spi_remove(struct spi_device *spi)
{
	struct lc89075wa_driver_data  *ddata = spi_get_drvdata(spi);
	struct intr_notify_list *st_notify;
	struct list_head *list;

#ifdef CONFIG_HAS_EARLYSUSPEND
	unregister_early_suspend(&ddata->es);
#endif

	mt_eint_registration(CUST_EINT_SPI_DIR_NUM, CUST_EINT_SPI_DIR_TYPE,
		NULL, 0);

	mutex_lock(&ddata->intr_notify_lock);
	list_for_each(list, &intr_notify_list_head.list) {
		st_notify = list_entry(list, struct intr_notify_list, list);
		list_del(list);
		kfree(st_notify);
	}
	mutex_unlock(&ddata->intr_notify_lock);
	kthread_stop(ddata->kthread_task_read_intr_status);

	mutex_destroy(&ddata->intr_notify_lock);
	mutex_destroy(&ddata->notify_regs_lock);

	/* make sure ops on existing fds can abort cleanly */
	spin_lock_irq(&ddata->spi_lock);
	spi_set_drvdata(spi, NULL);
	spin_unlock_irq(&ddata->spi_lock);
	mutex_destroy(&ddata->buf_lock);

	kfree(ddata);
	g_ddata = NULL;
	return 0;
}

static int lc89075wa_dir_spi_suspend(
	struct spi_device *spi, pm_message_t mesg)
{
	int retval = -EINVAL;

	dev_dbg(&spi->dev, "spi:lc89075 suspend\n");

	if (g_ddata && g_ddata->has_es)
		return 0;  /* Already done in early_suspend */

	mt_eint_registration(CUST_EINT_SPI_DIR_NUM, CUST_EINT_SPI_DIR_TYPE,
		NULL, 0);

	if (g_ddata) {
		g_ddata->suspend_flag = 1;  /* set flag */
		retval = lc89075wa_dir_send_command_multi(g_ddata->spi,
			LC89075WA_DIR_MODE_READ, 0,
			NULL, g_ddata->register_save_vals,
			sizeof(g_ddata->register_save_vals));
		if (!retval)
			g_ddata->register_saved = 1;
	}

	mt_set_gpio_out(GPIO_DIR_RST_EN_PIN, GPIO_OUT_ZERO);

	return 0;
}

static int lc89075wa_dir_spi_resume(struct spi_device *spi)
{
	struct lc89075wa_dir_platform_data *pfdata;
	int retval = -EINVAL;

	dev_dbg(&spi->dev, "spi:lc89075 resume\n");

	if (g_ddata && g_ddata->has_es)
		return 0;  /* Already done in late_resume */

	mt_set_gpio_out(GPIO_DIR_RST_EN_PIN, GPIO_OUT_ONE);

	/* wait until DIR got stable */
	usleep_range(LC89075WA_DIR_XMODE_STABLE_TIME,
		LC89075WA_DIR_XMODE_STABLE_TIME + 10);

	if (g_ddata) {
		g_ddata->suspend_flag = 0;  /* clear flag */
		if (!g_ddata->register_saved) {
			/* use default values */
			pfdata = (struct lc89075wa_dir_platform_data *)
				g_ddata->spi->dev.platform_data;
			memcpy(g_ddata->register_save_vals,
				pfdata->reg_init_value,
				sizeof(g_ddata->register_save_vals));
		}
		retval = lc89075wa_dir_send_command_multi(g_ddata->spi,
			LC89075WA_DIR_MODE_WRITE, 0,
			g_ddata->register_save_vals, NULL,
			sizeof(g_ddata->register_save_vals));
		if (!retval)
			g_ddata->register_saved = 0;
	}

	mt_eint_set_hw_debounce(
		CUST_EINT_SPI_DIR_NUM, CUST_EINT_SPI_DIR_DEBOUNCE_CN);
	mt_eint_registration(CUST_EINT_SPI_DIR_NUM, CUST_EINT_SPI_DIR_TYPE,
		lc89075wa_dir_irq_handler, 0);

	return 0;
}


struct spi_device_id lc89075wa_dir_id_table = {"spi_lc89075wa_dir", 0};

static struct spi_driver lc89075wa_dir_spi_driver = {
	.driver = {
		.name  = "spi_lc89075wa_dir",
		.bus   = &spi_bus_type,
		.owner = THIS_MODULE,
	},
	.probe    = lc89075wa_dir_spi_probe,
	.remove   = lc89075wa_dir_spi_remove,
	.suspend  = lc89075wa_dir_spi_suspend,
	.resume   = lc89075wa_dir_spi_resume,
	.id_table = &lc89075wa_dir_id_table,
};

const struct file_operations lc89075wa_dir_fops = {
	.owner          = THIS_MODULE,
	.open           = lc89075wa_dir_open,
	.release        = lc89075wa_dir_close,
	.read           = lc89075wa_dir_read,
	.write          = lc89075wa_dir_write,
	.poll           = lc89075wa_dir_poll,
	.unlocked_ioctl = lc89075wa_dir_ioctl,
};

static int __init lc89075wa_dir_init(void)
{
	int err;

	pr_info("%s:init\n", LC89075WA_DIR_DEV_NAME);

	lc89075wa_dir_chr_dev =
		kmalloc(sizeof(struct lc89075wa_dir_chr_dev_t), GFP_KERNEL);
	if (!lc89075wa_dir_chr_dev) {
		pr_err("kmalloc err\n");
		return -ENOMEM;
	}

	err = alloc_chrdev_region(&lc89075wa_dir_chr_dev->id, 0, 1,
		LC89075WA_DIR_DEV_NAME);
	if (err) {
		pr_err("alloc_chrdev_region err %d\n", err);
		goto out_free;
	}

	cdev_init(&lc89075wa_dir_chr_dev->cdev, &lc89075wa_dir_fops);
	lc89075wa_dir_chr_dev->cdev.owner = THIS_MODULE;

	err = cdev_add(&lc89075wa_dir_chr_dev->cdev,
		lc89075wa_dir_chr_dev->id, 1);
	if (err) {
		pr_err("cdev_add err %d\n", err);
		goto out_unregister;
	}

	lc89075wa_dir_chr_dev->class =
		class_create(THIS_MODULE, LC89075WA_DIR_DEV_NAME);
	if (IS_ERR(lc89075wa_dir_chr_dev->class)) {
		err = PTR_ERR(lc89075wa_dir_chr_dev->class);
		pr_err("class_create err %d\n", err);
		goto out_cdev_del;
	}

	lc89075wa_dir_chr_dev->dev =
		device_create(lc89075wa_dir_chr_dev->class, NULL,
			lc89075wa_dir_chr_dev->id, NULL,
			LC89075WA_DIR_DEV_NAME);
	if (IS_ERR(lc89075wa_dir_chr_dev->dev)) {
		err = PTR_ERR(lc89075wa_dir_chr_dev->dev);
		pr_err("device_create err %d\n", err);
		goto out_class_destroy;
	}

	/* spi_register_board_info() is called from model_init() */
	err = spi_register_driver(&lc89075wa_dir_spi_driver);
	if (err < 0) {
		pr_err("spi_device_register(lc89075wa_dir.c) err %d\n", err);
		goto out_device_destroy;
	}

	return 0;

out_device_destroy:
	device_destroy(lc89075wa_dir_chr_dev->class,
		lc89075wa_dir_chr_dev->id);

out_class_destroy:
	class_destroy(lc89075wa_dir_chr_dev->class);

out_cdev_del:
	cdev_del(&lc89075wa_dir_chr_dev->cdev);

out_unregister:
	unregister_chrdev_region(lc89075wa_dir_chr_dev->id, 1);

out_free:
	kfree(lc89075wa_dir_chr_dev);

	return err;
}

static void __exit lc89075wa_dir_exit(void)
{
	pr_info("lc89075wa driver exit\n");

	spi_unregister_driver(&lc89075wa_dir_spi_driver);

	device_destroy(lc89075wa_dir_chr_dev->class,
		lc89075wa_dir_chr_dev->id);
	class_destroy(lc89075wa_dir_chr_dev->class);
	cdev_del(&lc89075wa_dir_chr_dev->cdev);
	unregister_chrdev_region(lc89075wa_dir_chr_dev->id, 1);
	kfree(lc89075wa_dir_chr_dev);
}

module_init(lc89075wa_dir_init);
module_exit(lc89075wa_dir_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("kernel module lc89075wa_dir");
